/* lu.c
   lumines clone

Copyright 2005 Damian Yerrick

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Author contact information is in readme.txt

*/

#include "util.h"


/* Developer options */
#define PROFILE_GAME_ENGINE 0
#define SELECT_SWITCH_SKIN  0

/* Behavior options */
#define HALF_DELETES 1  /* Lumines = 1 */

/* one delete is worth this many points */
#define LU_DELETE_SCORE 40
/* if deletes make the field empty, grant this many points */
#define LU_EMPTY_SCORE 10000
/* if deletes make the field all one color, grant this many points */
#define LU_MONOCHROME_SCORE 1000

/* Speed of falling block in tiles per frame is
   (INITIAL_PIECE_SPEED + n_pieces * PIECE_SPEEDUP)/65536
   per frame. The INITIAL_PIECE_SPEED has been verified against Lumines. */
#define INITIAL_PIECE_SPEED 2100

/* Builds from 20050718 to 20050904 used 20 here for a difficulty
   curve more similar to Tetris, which becomes a twitch game above
   level 15.  Build 20050912 tried a slower value of 2 to approximate
   the slower, more "cerebral" pace of Lumines, but it turned out to
   be too easy.
   Apparently, Mr. Mizuguchi's reasoning was that if you want a
   twitch game, play the 1:00, 3:00, or 5:00 time trial. */
#define PIECE_SPEEDUP 4



/*

76543210
||||||||
|||||||+- Block is SE part of BLAWK
||||||+-- Block is SW part of BLAWK
|||||+--- Block is NE part of BLAWK
||||+---- Block is NW part of BLAWK
|||+----- Not used
||+------ Block contains bomb
++------- 0: empty space; 1: swept; 2: H block; 3: R block

*/
#define LU_SWEPT      0x40
#define LU_COLOR_H    0x80
#define LU_COLOR_R    0xC0
#define LU_COLOR_MASK (LU_COLOR_H | LU_COLOR_R)

#define LU_FUSE       0x10

#define LU_BOMB       0x20

#define LU_CONN_NW    0x08
#define LU_CONN_NE    0x04
#define LU_CONN_SW    0x02
#define LU_CONN_SE    0x01
#define LU_CONN_MASK  (LU_CONN_NW | LU_CONN_NE | LU_CONN_SW | LU_CONN_SE)

#define LU_FIELD_WID 16  /* size of playfield */
#define LU_FIELD_HT 10
#define LU_MAX_BOMBS 4
#define LU_MAP 0

#define LU_SPR_VRAM_H 0
#define LU_SPR_VRAM_R 4
#define LU_SPR_VRAM_BOMB 8
#define LU_SPR_VRAM_SWEEPER 16
#define LU_SPR_VRAM_DIGITS 64

#define g game_globals.left



enum {
  LU_INACTIVE,
  LU_NEW_GAME,
  LU_NEW_PIECE,
  LU_FALLING_PIECE,
  LU_GAMEOVER
};


/* g.skin_transition */
enum {
  LU_SKIN_LOAD_FILE,
  LU_SKIN_FADEOUT,
  LU_SKIN_LOAD_BG,
  LU_SKIN_LOAD_PF,
  LU_SKIN_LOAD_SPRITES,
  LU_SKIN_WAIT_SONG_END,
  LU_SKIN_LOAD_SONG,
  LU_SKIN_START_SONG,
  LU_SKIN_IDLE
};


#define LU_SCR_TOP    0  /* position and size of display */
#define LU_SCR_LEFT   4

#define LU_NEXT_PIECE_PITCH 32  /* y distance between next pieces */


#define GB_PERIOD(x) (2048 - (int)(131072 / (x)))

#define FREQ_BLAWK 77.78
#define FREQ_ROTATE 1244.51
#define FREQ_MOVE (FREQ_ROTATE * 1.5)


static u32 pal_save[16];

static
void lu_next_skin(void);


#if PROFILE_GAME_ENGINE
#define PROFILE_RESET profiling_y = 0
#define PROFILE_START profiling_y -= LCD_Y
#define PROFILE_END   profiling_y += LCD_Y
static unsigned int profiling_y = 0;
#else
#define PROFILE_RESET (void)0
#define PROFILE_START (void)0
#define PROFILE_END   (void)0
#endif


static
void decimalize_spr(signed int x, signed int y, unsigned int n)
{
  if(x <= -8 || x >= 320 || y <= -8 || y >= 160)
    return;

  y &= 0x00ff;
  do {
    unsigned int tens = dv(n, 10);
    
    u.oam[u.n_sprites].y = y | OAM_SQUARE;
    u.oam[u.n_sprites].x = (x & 0x1ff) | OAM_SIZE1;
    u.oam[u.n_sprites].tile = 0xc000 | (LU_SPR_VRAM_DIGITS + (n - tens * 10) * 4);
    u.n_sprites++;
    n = tens;
    x -= 10;
  } while(n > 0 && x > -8);
}



/* lu_wait4vbl() ****************
   Waits for vblank, then sets up the hdma for the 12x12 grid.
*/
static
void lu_wait4vbl(int hide)
{
#if PROFILE_GAME_ENGINE
  static unsigned char last_y[16];
  unsigned int max_y = 0;
#endif
  unsigned int y;
  unsigned int playing_music = !hide && (u.state != LU_GAMEOVER);

#if PROFILE_GAME_ENGINE
  y = LCD_Y;
  if(y < 160)
    y += 228;
  y -= 160;
  decimalize_spr(230, 124, y);
  last_y[u.retraces & 0x0f] = y;
  for(y = 0; y < 16; y++)
    if(max_y < last_y[y])
      max_y = last_y[y];
  decimalize_spr(230, 148, profiling_y);
  decimalize_spr(230, 136, max_y);
#endif

  if(playing_music)
  {
    if(g.skin_transition == LU_SKIN_START_SONG)
    {
      gsm_playback(u.cur_song.name);
      g.skin_transition = LU_SKIN_IDLE;
    }
    else
    {
      gsm_playback(0);
    }
  }
  for(y = u.n_sprites; y < GBA_NSPRITES; y++)
  {
    u.oam[y].y = OAM_HIDDEN;
  }
  u.n_sprites = 0;
  u.retraces++;
  wait4vbl();
  if(playing_music)
    dsound_switch_in();
  else
    DMA[1].control = 0;  /* silence GSM playback */
  DMA[0].control = 0;
  if(hide)
  {
    LCDMODE = 0 | LCDMODE_BG2 | LCDMODE_SPR | LCDMODE_1DSPR;
  }
  else
  {
    BGSCROLL[0] = g.hdma[0][0];
    BGSCROLL[1] = g.hdma[0][1];
    DMA[0].src = &(g.hdma[1]);
    DMA[0].dst = (u16 *)&(BGSCROLL[0]);
    DMA[0].count = 2;
    DMA[0].control = DMA_DSTINCRESET | DMA_SRCINC | DMA_REPEAT | DMA_U32 |
                     DMA_HBLANK | DMA_ENABLE;
    LCDMODE = 0 | LCDMODE_BG0 | LCDMODE_BG1 | LCDMODE_BG2 | LCDMODE_SPR | LCDMODE_1DSPR;
    lu_next_skin();
  }
  dma_memcpy((void *)OAM, u.oam, sizeof(u.oam));
  make_repeats(~JOY);
}


static
void gba_draw_sweeper(void)
{
  unsigned int x, y;
  x = 8 * LU_SCR_LEFT + 12 * g.sweep_x + ((12 * u.sweep_sub) >> 24) - 4;
  

  /* Draw the sweeper */
  for(y = 0; y < 5; y++)
  {
    u.oam[u.n_sprites].y = (32 + 24 * y) | OAM_TALL;
    u.oam[u.n_sprites].x = x | OAM_SIZE2;
    u.oam[u.n_sprites].tile = 0x9010;
    u.n_sprites++;
  }

  /* Draw the delete count */
  decimalize_spr(x + 3, 21, g.deletes_this_pass);
  if(g.deletes_move_cd)
  {
    decimalize_spr(222,
                   (7 * g.deletes_move_cd / 16) + 11,
                   g.deletes_last_pass);
  }
}


static
void gba_draw_score(void)
{
  decimalize_spr(222, 11, g.deletes);
  decimalize_spr(82, 11, g.score);
}


static void gba_draw_score_add(const LuScoreAdd *sc)
{
  unsigned int countdown = sc->countdown;
  signed int t = countdown > 16 ? 16 : countdown;
  signed int x = 68 + (((sc->start_x - 68) * t) >> 4);
  signed int y = 11 + (((sc->start_y - 11) * t) >> 4);

  decimalize_spr(x, y, sc->value);
}


static
void handle_bonuses(void)
{
  unsigned int i;

  for(i = 0; i < LU_SCORE_ADD_N; i++)
  {
    unsigned int t = g.score_add[i].countdown;

    if(t > 0)
    {
      gba_draw_score_add(&(g.score_add[i]));
      g.score_add[i].countdown = t - 1;
      if(t == 1)
        g.score += g.score_add[i].value;
    }
  }
}


static
void add_score(unsigned int x,
               unsigned int y,
               unsigned int countdown,
               unsigned int value)
{
  unsigned int i;

  for(i = 0; i < 4; i++)
  {
    if(g.score_add[i].countdown == 0)
    {
      g.score_add[i].start_x = x;
      g.score_add[i].start_y = y;
      g.score_add[i].countdown = countdown;
      g.score_add[i].value = value;
      return;
    }
  }
}


#if 0
static
void gba_debug_draw_columns(unsigned int c)
{
  unsigned int x;

  for(x = 0; x < 16; x++, c >>= 1)
  {
    if(!(c & 1))
      continue;
    u.oam[u.n_sprites].y = 152 | OAM_SQUARE;
    u.oam[u.n_sprites].x = (32 + 12 * x) | OAM_SIZE1;
    u.oam[u.n_sprites].tile = 0xA000;
    u.n_sprites++;
  }
}
#endif


/* gba_draw_field() ********************
   Draws columns of the playfield that are marked dirty.
*/
static
void gba_draw_field(void)
{
  unsigned int x, y;
  unsigned int cols = g.dirty_columns;

  if(!cols)
    return;

  /* Draw everything so i don't have to draw anything next time */
  g.dirty_columns = 0;

  for(x = 0; x < LU_FIELD_WID; x++, cols >>= 1)
  {
    if(1 & cols)  /* redraw only dirty columns */
    {
      u16 *map_base = MAP[LU_MAP + (x & 1)][LU_SCR_TOP]
                      + LU_SCR_LEFT + 3 * (x >> 1);
      unsigned int j;

      /* remove cached positions of bombs */
      for(j = 0; j < LU_MAX_BOMBS; j++)
        if(g.destroyer_x[j] == x)
          g.destroyer_x[j] = LU_FIELD_WID;

      for(y = 0; y < LU_FIELD_HT; y++, map_base += 64)
      {
        unsigned int blk = g.b[y][x];
        unsigned int attr = ((blk & LU_COLOR_MASK) << 6)
                            | (blk & LU_SWEPT)
                            | 0x8000;

        if(!blk)
          attr = 32*4;
        else if((blk & LU_COLOR_MASK) == LU_SWEPT)
          attr = 0x9084;
        else if(blk & LU_FUSE)
          attr |= 0x4000;
        attr |= (blk & LU_CONN_MASK) << 2;
        

        *(map_base     ) = attr;
        *(map_base +  1) = attr + 1;
        *(map_base + 32) = attr + 2;
        *(map_base + 33) = attr + 3;

        /* if this is a bomb, cache it */
        if(blk & LU_BOMB)
          for(j = 0; j < LU_MAX_BOMBS; j++)
            if(g.destroyer_x[j] >= LU_FIELD_WID)
            {
              g.destroyer_x[j] = x;
              g.destroyer_y[j] = y;
              break;
            }
      }
    }
  }
}


/* gba_draw_piece_at() ***************
   10
   23
*/
static
void gba_draw_piece_at(unsigned int x, unsigned int y, unsigned int colors)
{
  static const unsigned char x_offset[4] = {12, 0, 0, 12};
  static const unsigned char y_offset[4] = {0, 0, 12, 12};
  unsigned int i;

  for(i = 0; i < 4; i++, colors >>= 2)
  {
    if(colors & 1)
    {
      u.oam[u.n_sprites].y = ((y + y_offset[i]) & 0x0FF) | OAM_SQUARE;
      u.oam[u.n_sprites].x = ((x + x_offset[i]) & 0x1FF) | OAM_SIZE1;
      u.oam[u.n_sprites].tile = 0x8000 | LU_SPR_VRAM_BOMB;
      u.n_sprites++;
    }
    u.oam[u.n_sprites].y = ((y + y_offset[i]) & 0x0FF) | OAM_SQUARE;
    u.oam[u.n_sprites].x = ((x + x_offset[i]) & 0x1FF) | OAM_SIZE1;
    u.oam[u.n_sprites].tile = ((colors & 2) << 11)
                              ? 0xB000 | LU_SPR_VRAM_R
                              : 0xA000 | LU_SPR_VRAM_H;
    u.n_sprites++;
  }
}


static
void gba_draw_pieces(void)
{
  unsigned int i;
  unsigned int new_piece_phase = (u.state == LU_NEW_PIECE)
                                 ? u.state_time : 0;
  unsigned int next_off = new_piece_phase * LU_NEXT_PIECE_PITCH / 16;
  unsigned int x_dst = (LU_SCR_LEFT * 8 + g.x * 12);

  if(u.state == LU_NEW_PIECE)
  {
    g.x_pixel = x_dst;
    x_dst = (16 - new_piece_phase) * x_dst / 16;
  }
  else if(u.state == LU_FALLING_PIECE)
  {
    x_dst = (g.x_pixel + x_dst + (x_dst > g.x_pixel)) >> 1;
    g.x_pixel = x_dst;
  }

  if(u.state == LU_NEW_PIECE || u.state == LU_FALLING_PIECE)
  gba_draw_piece_at(x_dst,
                    ((g.y * 12) >> 16) + 32,
                    g.piece[0]);

  gba_draw_piece_at(0,         8 + next_off,                       g.piece[1]);
  gba_draw_piece_at(0,         8 + LU_NEXT_PIECE_PITCH + next_off, g.piece[2]);
  gba_draw_piece_at(-next_off, 8 + LU_NEXT_PIECE_PITCH * 2,        g.piece[3]);

  /* draw bombs */
  for(i = 0; i < LU_MAX_BOMBS; i++)
  {
    unsigned int x = g.destroyer_x[i], y = g.destroyer_y[i];

    if(x < LU_FIELD_WID)
    {
      u.oam[u.n_sprites].y = (y * 12 + 32) | OAM_SQUARE;
      u.oam[u.n_sprites].x = (x * 12 + LU_SCR_LEFT * 8) | OAM_SIZE1;
      u.oam[u.n_sprites].tile = 0x8000 | LU_SPR_VRAM_BOMB;
      u.n_sprites++;
    }
  }
}


static
void unpack_12x12_tile(u32 *dst, const u16 *src)
{
  unsigned int h;

  for(h = 8; h > 0; h--, src += 3, dst++)
  {
    unsigned int a = src[0];
    unsigned int b = src[1];
    unsigned int c = src[2];
    dst[0] = a | (b << 16);
    dst[8] = c;
  }
  dst += 8;
  for(h = 4; h > 0; h--, src += 3, dst++)
  {
    unsigned int a = src[0];
    unsigned int b = src[1];
    unsigned int c = src[2];
    dst[0] = a | (b << 16);
    dst[8] = c;
  }
  for(h = 4; h > 0; h--, src += 3, dst++)
  {
    dst[0] = 0;
    dst[8] = 0;
  }
}


static
void unpack_12x24_tile(u32 *dst, const u16 *src)
{
  unsigned int h, tile_h = 3;

  for(; tile_h > 0; tile_h--)
  {
    for(h = 8; h > 0; h--, src += 3, dst++)
    {
      unsigned int a = src[0];
      unsigned int b = src[1];
      unsigned int c = src[2];
      dst[0] = a | (b << 16);
      dst[8] = c;
    }
    dst += 8;
  }
}


static
void lu_next_skin(void)
{
  unsigned int y;
  const unsigned char *src = u.cur_skin_data;

  switch(g.skin_transition)
  {
    case LU_SKIN_LOAD_FILE:
      if(++g.cur_skin >= gbfs_count_objs(Lskins_gbfs))
        g.cur_skin = 0;
      u.cur_skin_data = gbfs_get_nth_obj(Lskins_gbfs, g.cur_skin, NULL, NULL);
      /* if it's a new game (0 pieces dropped), skip the fadeout */
      g.skin_transition = g.n_pieces ? LU_SKIN_FADEOUT : LU_SKIN_LOAD_BG;
      break;

    case LU_SKIN_FADEOUT:
      set_palette_dimmed(PALRAM+0x20, pal_save,     RGB( 0, 0, 0), g.sweep_x + 1);
      set_palette_dimmed(PALRAM+0x30, pal_save + 8, RGB( 0, 0, 0),  g.sweep_x + 1);
      PALRAM[0] = PALRAM[0x20];
      u.gsm_volume = 248 - 16 * g.sweep_x - (u.sweep_sub >> 20);
      if(g.sweep_x == 15)
        g.skin_transition = LU_SKIN_LOAD_BG;
      break;

    case LU_SKIN_LOAD_BG:
      /* set background */
      dma_memcpy(PATRAM4(2, 384), src + 256, 19200);
      dma_memcpy(pal_save, src, 64);  /* load bg palette, but not into VRAM yet */
      g.skin_transition = LU_SKIN_LOAD_PF;
      PALRAM[0] = PALRAM[0x20];
      break;

    case LU_SKIN_LOAD_PF:
      dma_memcpy(PALRAM + 0x0A0, src + 64, 64);  /* h and r, to bg */
      dma_memcpy(PALRAM + 0x180, src + 128, 32);  /* bomb */
      dma_memcpy(PALRAM + 0x090, src + 160, 32);  /* swept */
      dma_memcpy(PALRAM + 0x190, src + 192, 32);  /* sweeper */
      dma_memcpy(PALRAM + 0x1c0, src + 224, 32);  /* digits */
      src += 256 + 19200;
      for(y = 0; y < 34; y++, src += 72)
        unpack_12x12_tile(PATRAM4(2, y * 4), (const u16 *)src);

      unpack_12x24_tile(SPR_VRAM(LU_SPR_VRAM_SWEEPER), (const u16 *)src);
      src += 144 + 72 + 72;
      for(y = 0; y < 10; y++, src += 72)
        unpack_12x12_tile(SPR_VRAM(LU_SPR_VRAM_DIGITS + y * 4), (const u16 *)src);

      /* pull bomb tile into sprite cel VRAM now and erase it from PF */
      dma_memcpy(SPR_VRAM(LU_SPR_VRAM_BOMB), PATRAM4(2, 32 * 4), 4*32);  /* bomb */
      for(y = 0; y < 32; y++)
        PATRAM4(2, 32*4)[y] = 0;

      g.skin_transition = LU_SKIN_LOAD_SPRITES;
      break;

    case LU_SKIN_LOAD_SPRITES:
      /* pull various tiles into sprite cel VRAM */
      dma_memcpy(SPR_VRAM(LU_SPR_VRAM_H), PATRAM4(2, 0 * 4), 4*32);  /* lone H block */
      dma_memcpy(SPR_VRAM(LU_SPR_VRAM_R), PATRAM4(2, 16 * 4), 4*32);  /* lone R block */

      dma_memcpy(PALRAM + 0x1A0, src + 64, 64);  /* h and r, to sprite */

      g.skin_transition = LU_SKIN_WAIT_SONG_END;
      break;

    case LU_SKIN_WAIT_SONG_END:
      break;

    case LU_SKIN_LOAD_SONG:
      dma_memcpy(PALRAM + 0x020, pal_save, 64);  /* load bg palette */
      src += 256 + 19200 + 72*48;
      /* set song needed */
      dma_memcpy(&u.cur_song, src, sizeof(u.cur_song));
      src += sizeof(u.cur_song);
      g.skin_transition = LU_SKIN_START_SONG;
      PALRAM[0] = PALRAM[0x20];
      break;

  }
}


static
void init_vram(void)
{
  static const unsigned char pause_pal[] = { 0, 1, 2, 3 };
  unsigned int len;
  const void *src;

  LCDMODE = 0;
  PALRAM[0] = 0;

  src = gbfs_get_obj(data_gbfs, "pause.chr", &len);
  expand_gb_to_gba4bit(SPR_VRAM(896), src, len, pause_pal);

  while(LCD_Y != 161) ;

  BGCTRL[0] = BGCTRL_PAT(2) | BGCTRL_16C | BGCTRL_NAME(LU_MAP)
              | BGCTRL_H32 | BGCTRL_V32;
  BGCTRL[1] = BGCTRL_PAT(2) | BGCTRL_16C | BGCTRL_NAME(LU_MAP + 1)
              | BGCTRL_H32 | BGCTRL_V32;
  BGCTRL[2] = BGCTRL_PAT(2) | BGCTRL_16C | BGCTRL_NAME(31)
              | BGCTRL_H32 | BGCTRL_V32;
  BGSCROLL[0].x = 0;
  BGSCROLL[0].y = 0;
  BGSCROLL[1].x = 512 - 12;
  BGSCROLL[1].y = 0;
  BGSCROLL[2].x = 0;
  BGSCROLL[2].y = 256 - 160;

  for(len = 0; len < 1024; len++)
    MAP[LU_MAP][0][len] = 128;
  for(len = 0; len < 1024; len++)
    MAP[LU_MAP + 1][0][len] = 128;

  /* Generate background map */
  len = 384;
  {
    unsigned int x, y;
    for(y = 0; y < 20; y++)
      for(x = 0; x < 30; x++)
      {
        unsigned int pal = (y >= 4 && y < 19 && x >= 4 && x < 28)
                            ? 0x3000 : 0x2000;
        MAP[31][y + 12][x] = len++ | pal;
      }
  }

  /* Generate HDMA stuff for 12-pixel-tall rows */
  {
    int pos = -32;
    int twelveness = -32;
    for(len = 0; len < 160; len++)
    {
      g.hdma[len][0].x = 0;
      g.hdma[len][0].y = pos;
      g.hdma[len][1].x = 512 - 12;
      g.hdma[len][1].y = pos;

      if(++twelveness >= 12)
      {
        twelveness = 0;
        pos += 4;
      }
    }
  }

  u.n_sprites = 0;

}


/* steps to bomb
   Adapt flood fill from ../tod/tetanus.c
   sweep_column() should score for fuses
*/

static
void fuse_flood_fill(int x, int y)
{
  int fillL, fillR, i;
  int still_in = 1;
  unsigned int fillC = g.b[y][x];

  if(fillC & LU_FUSE)
    return;
  fillC &= LU_COLOR_MASK;

  decimalize_spr(24, 160-12*4, x);
  decimalize_spr(24, 160-12*3, y);

  fillL = fillR = x;
  while(still_in)
  {
    g.b[y][fillL] |= LU_FUSE;
    g.dirty_columns |= 1 << fillL;
    fillL--;
    still_in = (fillL < 0 || (1 & (g.floaty_columns >> fillL)))
               ? 0
               : ((g.b[y][fillL] & (LU_COLOR_MASK | LU_FUSE)) == fillC);
  }
  fillL++;

  still_in = 1;
  while(still_in)
  {
    g.b[y][fillR] |= LU_FUSE;
    g.dirty_columns |= 1 << fillR;
    fillR++;
    still_in = (fillR > LU_FIELD_WID - 1 || (1 & (g.floaty_columns >> fillL)))
               ? 0
               : ((g.b[y][fillR] & (LU_COLOR_MASK | LU_FUSE)) == fillC);
  }
  fillR--;

  if(y > 0)
    for(i = fillL; i <= fillR; i++)
      if((g.b[y - 1][i] & (LU_COLOR_MASK | LU_FUSE)) == fillC)
	fuse_flood_fill(i, y - 1);

  if(y + 1 < LU_FIELD_HT)
    for(i = fillL; i <= fillR; i++)
      if((g.b[y + 1][i] & (LU_COLOR_MASK | LU_FUSE)) == fillC)
	fuse_flood_fill(i, y + 1);
}


/* mark_blawks() ***************
   Look for 2x2 squares of the same color and mark them as BLAWKs.
   Also start a fuse at any bomb within a BLAWK.
*/
static
unsigned int mark_blawks()
{
  /* Don't mark BLAWKs that include columns that haven't fallen yet */
  unsigned int cols = ~g.floaty_columns
                      & ~(g.floaty_columns >> 1);
  unsigned int x, y;
  unsigned int n_found = 0;

  /* Strange: In Lumines, if you make a BLAWK whose left side overlaps
     the sweeper, it'll sweep the right half of the BLAWK and turn the
     left half of that BLAWK back into normal blocks. The code that
     used to prevent */
#if !HALF_DELETES
  /* skip columns too close to the sweeper */
  cols &= ~(1 << g.sweep_x);
#endif

  for(x = 0; x < LU_FIELD_WID - 1; x++)
  {
    if(!(cols & (1 << x)))
      continue;
    for(y = 0; y < LU_FIELD_HT - 1; y++)
    {
      unsigned int blk = g.b[y][x];
      unsigned int color = blk & LU_COLOR_MASK;

      if(!blk)  /* skip blank space */
        continue;
      if(color == LU_SWEPT)  /* skip already swept blocks */
        continue;
      if(blk & LU_CONN_NW)  /* skip already formed BLAWKs */
        continue;
      /* skip areas that are swept or aren't BLAWKs */
      if(color != (g.b[y    ][x + 1] & LU_COLOR_MASK))
        continue;
      if(color != (g.b[y + 1][x    ] & LU_COLOR_MASK))
        continue;
      if(color != (g.b[y + 1][x + 1] & LU_COLOR_MASK))
        continue;

      /* by now we've verified that this is the upper left corner
         of a newly formed BLAWK */
      g.b[y    ][x    ] |= LU_CONN_NW;
      g.b[y    ][x + 1] |= LU_CONN_NE;
      g.b[y + 1][x    ] |= LU_CONN_SW;
      g.b[y + 1][x + 1] |= LU_CONN_SE;

      /* if it contains a bomb, set a fuse */
      if((g.b[y][x] | g.b[y][x + 1] | g.b[y + 1][x    ] | g.b[y + 1][x + 1]) & LU_BOMB)
        fuse_flood_fill(x, y);
      n_found++;
      g.dirty_columns |= 3 << x;
    }
  }

  if(n_found)
  {
    SQR2CTRL = 0xf400;
    SQR2FREQ = FREQ_RESET | u.cur_song.blawk_period;
  }
  return n_found;
}


/* unmark_blawks_in()
   From a given column, removes the BLAWKs above blank space.
   Called on a column from which tiles have been removed.
*/
static
void unmark_blawks_in(unsigned int x)
{
  /* The states of space_seen:
     0  Have not seen a blank space
     1  Have seen a blank space with no tile above it
     2  Have seen a blank space with one or more tiles above it
  */
  unsigned int space_seen = 0;
  unsigned int y;
  for(y = LU_FIELD_HT - 1; y > 0; y--)
  {
    unsigned int blk = g.b[y][x];

    if(!blk)
    {
      if(space_seen == 0)
        space_seen = 1;
    }
    else if(space_seen == 1)
    {
      space_seen = 2;
    }
    if((blk & LU_CONN_SE) && space_seen)
    {
      g.b[y    ][x    ] &= ~LU_CONN_SE;
      g.b[y    ][x - 1] &= ~LU_CONN_SW;
      g.b[y - 1][x    ] &= ~LU_CONN_NE;
      g.b[y - 1][x - 1] &= ~LU_CONN_NW;
      g.dirty_columns |= 3 << (x - 1);
    }
    if((blk & LU_CONN_SW) && space_seen)
    {
      g.b[y    ][x + 1] &= ~LU_CONN_SE;
      g.b[y    ][x    ] &= ~LU_CONN_SW;
      g.b[y - 1][x + 1] &= ~LU_CONN_NE;
      g.b[y - 1][x    ] &= ~LU_CONN_NW;
      g.dirty_columns |= 3 << x;
    }
  }

  /* Make sure that a lone floating tile at the very top
     of the playfield gets marked for falling. */
  if(space_seen == 1 && g.b[0][x])
    space_seen = 2;

  /* If tiles are floating, drop them. */
  if(space_seen == 2)
    g.floaty_columns |= 1 << x;
}


/* unmark_fuses() *******************
   Removes the fuse from all blocks.  Called after setting up gravity.
 */
static
void unmark_fuses(void)
{
  unsigned int x, y;

  for(y = 0; y < LU_FIELD_HT; y++)
    for(x = 0; x < LU_FIELD_WID; x++)
    {
      unsigned int c = g.b[y][x];
      if(c & LU_FUSE)
      {
        g.b[y][x] = c & ~LU_FUSE;
        g.dirty_columns |= 1 << x;
      }
    }
}


/* set_fuses() **********************
   Extends existing fuses.
*/
static
void set_fuses(void)
{
  unsigned int x, y;

  for(y = 0; y < LU_FIELD_HT; y++)
    for(x = 0; x < LU_FIELD_WID; x++)
    {
      unsigned int a = g.b[y][x];

      if((a & LU_CONN_MASK) && (a & LU_BOMB) && !(a & LU_FUSE))
        fuse_flood_fill(x, y);
    }

  /* extend fuses to left and right */
  for(y = 0; y < LU_FIELD_HT; y++)
    for(x = 0; x < LU_FIELD_WID - 1; x++)
    {
      unsigned int a = g.b[y][x], b = g.b[y][x + 1];
      if(!a || !b)
        continue;
      if((a & LU_COLOR_MASK) != (b & LU_COLOR_MASK))
        continue;
      if((a & LU_FUSE) && !(b & LU_FUSE))
        fuse_flood_fill(x + 1, y);
      else if((b & LU_FUSE) && !(a & LU_FUSE))
        fuse_flood_fill(x, y);
    }

  /* extend fuses up and down */
  for(y = 0; y < LU_FIELD_HT - 1; y++)
    for(x = 0; x < LU_FIELD_WID; x++)
    {
      unsigned int a = g.b[y][x], b = g.b[y + 1][x];
      if(!a || !b)
        continue;
      if((a & LU_COLOR_MASK) != (b & LU_COLOR_MASK))
        continue;
      if((a & LU_FUSE) && !(b & LU_FUSE))
        fuse_flood_fill(x, y + 1);
      else if((b & LU_FUSE) && !(a & LU_FUSE))
        fuse_flood_fill(x, y);
    }
}  


/* sweep_column() ********************
   Overwrites all BLAWKs within this column with swept blocks.
   For NW sides of blocks, adds one 
   Returns the number of swept blocks in this column.
   If only the right side of a BLAWK has been swept (e.g. BLAWK formed
   on top of sweeper), disconnects blocks to the left.
*/
static
unsigned int sweep_column(unsigned int x)
{
  unsigned int n = 0;
  unsigned int y;

  /* sweep up all blocks in this column */
  for(y = 0; y < LU_FIELD_HT - 1; y++)
  {
    unsigned int blk = g.b[y][x];

    /* Score for deletes */
    if(blk & LU_CONN_NW)
      g.sweepgrp_deletes += 1;

    /* De-BLAWK area to left */
    if((blk & LU_CONN_NE) && (x > 0))
    {
      if(g.b[y][x - 1] & LU_CONN_NW)
      {
        g.b[y ][x - 1] &= ~LU_CONN_NW;
        g.dirty_columns |= 1 << (x - 1);
      }
      if(g.b[y + 1][x - 1] & LU_CONN_SW)
      {
        g.b[y + 1][x - 1] &= ~LU_CONN_SW;
        g.dirty_columns |= 1 << (x - 1);
      }
    }
    if(blk & (LU_CONN_NW | LU_CONN_NE))
    {
      g.b[y    ][x    ] = LU_SWEPT;
      g.b[y + 1][x    ] = (g.b[y + 1][x    ] & (LU_CONN_NW | LU_CONN_NE)) | LU_SWEPT;
    }
    else if(blk & LU_FUSE)
    {
      g.b[y    ][x    ] = LU_SWEPT;
    }
  }

  if(g.b[y][x] & LU_FUSE)
    g.b[y    ][x    ] = LU_SWEPT;

  /* count swept blocks in this column */
  for(y = 0; y < LU_FIELD_HT; y++)
  {
    unsigned int blk = g.b[y][x];

    if((blk & LU_COLOR_MASK) == LU_SWEPT)
      n++;
  }

  if(n)
    g.dirty_columns |= 1 << x;
  return n;
}


/* delete_swept() **************
   Deletes swept blocks in [from_x, to_x).  Breaks up BLAWKs formed
   on top of swept blocks.  Breaks up fuses.
*/
static
int delete_swept(unsigned int from_x, unsigned int to_x)
{
  unsigned int x, y;
  unsigned int n_swept = 0;

  if(to_x > LU_FIELD_WID)
    to_x = LU_FIELD_WID;
  if(to_x <= from_x)
    return 0;

  g.deletes_this_pass += g.sweepgrp_deletes;
  g.sweepgrp_deletes = 0;
  for(x = from_x; x < to_x; x++)
  {
    unsigned int need_unmark = 0;
    for(y = 0; y < LU_FIELD_HT; y++)
    {
      unsigned int blk = g.b[y][x];

      if((blk & LU_COLOR_MASK) == LU_SWEPT)
      {
        g.b[y][x] = 0;
        g.dirty_columns |= 1 << x;
        need_unmark = 1;
        n_swept++;
      }
    }
    if(need_unmark)
      unmark_blawks_in(x);
  }
  if(n_swept)
  {
    /* Turn off fuses. They'll get turned back on once gravity ends. */
    unmark_fuses();

    /* If unmark_blawks_in() did not find any floating columns,
       mark one column as floating so that the gravity code has a
       chance to turn fuses back on. */
    if(!g.floaty_columns)
      g.floaty_columns = 1;
  }

  return n_swept;
}


static void check_monochrome(void)
{
  unsigned int x, y, h = 0, r = 0;

  for(y = 0; y < LU_FIELD_HT; y++)
    for(x = 0; x < LU_FIELD_WID; x++)
    {
      unsigned int b = g.b[y][x] & LU_COLOR_MASK;
      if(b == LU_COLOR_H)
        h++;
      else if(b == LU_COLOR_R)
        r++;
    }

  if(h == 0 && r == 0)
  {
    add_score(8 * LU_SCR_LEFT + 12 * g.sweep_x,
              32 + 12 * LU_FIELD_HT - 11,
              60,
              LU_EMPTY_SCORE);
    SQR2CTRL = 0xf400;
    SQR2FREQ = FREQ_RESET | (1024 + (u.cur_song.blawk_period >> 1));
  }
  else if(h == 0 || r == 0)
  {
    add_score(8 * LU_SCR_LEFT + 12 * g.sweep_x,
              32 + 12 * LU_FIELD_HT - 11,
              60,
              LU_MONOCHROME_SCORE);
    SQR2CTRL = 0xf400;
    SQR2FREQ = FREQ_RESET | (1024 + (u.cur_song.blawk_period >> 1));
  }
}


/* sweep_finish() ****************
   Finishes a swept group whose leftmost square is g.sweepgrp_left
   and whose rightmost square is g.sweep_x.
*/
static
void sweep_finish(void)
{
  if(g.sweepgrp_left < g.sweep_x)
  {
    int n_swept;

  PROFILE_RESET;
  PROFILE_START;
    n_swept = delete_swept(g.sweepgrp_left, g.sweep_x);

    if(n_swept)
      check_monochrome();
  PROFILE_END;
  }
}


/* rest_frames() ***************
   Draws sprites and rests for n frames.
   Stops and returns 1 if A was pressed.
*/
static
int rest_frames(unsigned int n)
{
  for(; n > 0; n--)
  {
    gba_draw_field();
    gba_draw_sweeper();
    gba_draw_pieces();
    gba_draw_score();
    lu_wait4vbl(0);
    if(u.repeats[0] == 1)
      return 1;
  }
  return 0;
}


static
void make_new_piece(void)
{
  unsigned int i;

  g.x = 7;
  g.y = -2 << 16;
  u.state_time = 16;
  u.state = LU_NEW_PIECE;

  for(i = 0; i < 3; i++)
    g.piece[i] = g.piece[i + 1];

  g.y_speed = INITIAL_PIECE_SPEED + g.n_pieces++ * PIECE_SPEEDUP;

  if(!(JOY & JOY_SELECT))
  {
    g.piece[3] = 0xa1;
    return;
  }

  i = rand();
  g.piece[3] = (i >> 7) & 0xaa;
  if(((i >> 6) & 0xab) == 0)
    g.piece[3] |= 1;
}


static
void new_game(void)
{
  unsigned int x, y;

  /* clear playfield */
  for(y = 0; y < LU_FIELD_HT; y++)
    for(x = 0; x < LU_FIELD_WID; x++)
      g.b[y][x] = 0;

  /* clear cached positions of bombs */
  for(y = 0; y < LU_MAX_BOMBS; y++)
    g.destroyer_x[y] = LU_FIELD_WID;

  /* clear score add */
  for(y = 0; y < LU_SCORE_ADD_N; y++)
    g.score_add[y].countdown = 0;

  /* Entire screen needs to be redrawn */
  g.dirty_columns = (1 << LU_FIELD_WID) - 1;
  g.floaty_columns = 0;

  /* lu_next_skin() uses g.n_pieces to discriminate between loading
     a skin at game start and loading a lu skin at game start */
  g.n_pieces = 0;
  g.skin_transition = LU_SKIN_LOAD_FILE;
  while(g.skin_transition < LU_SKIN_WAIT_SONG_END)
    lu_next_skin();
  g.skin_transition = LU_SKIN_LOAD_SONG;

  srand((rand() ^ u.retraces) + g.score);
  make_new_piece();
  make_new_piece();
  make_new_piece();

  g.sweep_x = 0;
  u.sweep_sub = 0;
  g.sweepgrp_left = 0;
  g.sweepgrp_deletes = 0;
  g.deletes = 0;
  g.deletes_last_pass = 0;
  g.deletes_this_pass = 0;
  g.score = 0;
  u.next_skin_at = 100;  /* this many blocks */

  /* Always start out with the same piece during debugging. */
  g.piece[0] = 0xa1;

}


static void game_over(void)
{
  u.state = LU_GAMEOVER;
  set_palette_dimmed(PALRAM+0x20, pal_save, RGB(31, 0, 0),  8);
  set_palette_dimmed(PALRAM+0x30, pal_save + 8, RGB(31, 0, 0),  8);
  PALRAM[0] = PALRAM[32];
}


static
int piece_overlaps(signed int x, signed int y)
{
  if(x < 0 || x > LU_FIELD_WID - 2 || y > 8)
    return 1;
  if(y < -1)
    return 0;
  if(g.b[y + 1][x])
    return 1;
  if(g.b[y + 1][x + 1])
    return 1;
  return 0;
}


static
void pf_gravity(void)
{
  u32 cols = g.floaty_columns;
  unsigned int x;
  signed int y;
  unsigned int hit_floor = 0;

  if(!cols)
    return;

  for(x = 0; x < LU_FIELD_WID; x++, cols >>= 1)
  {
    unsigned int got = 0;
    unsigned int in_floor = 1;

    if(!(cols & 1))
      continue;

    for(y = LU_FIELD_HT - 2; y >= 0; y--)
    {
      unsigned int top = g.b[y][x], bottom = g.b[y + 1][x];

      if(top && !bottom)
      {
        g.b[y + 1][x] = top & ~LU_CONN_MASK;
        g.b[y][x] = 0;
        got = 1;
        if(in_floor)
          hit_floor = 1;
      }
      if(!bottom)
        in_floor = 0;

      if(got)  /* if we moved at least one piece */
        cols |= 0x10000;  /* say it's still falling */
    }
  }

  /* mark the columns that were changed */
  g.floaty_columns = cols;
  g.dirty_columns |= cols;

  if(hit_floor && (u.state != LU_NEW_PIECE || u.state_time < 10))
  {
    NOISECTRL = 0xf100;
    NOISEFREQ = FREQ_RESET | 0x80;
  }

  if(!g.floaty_columns)
    set_fuses();
}


/* lock_piece() ****************
   Locks a piece into place.
   Returns 0 for OK or nonzero for game over.
   CORNER CASE CAUTION: Behavior close to ceiling is
   not tested against Lumines.
*/
static
int lock_piece(void)
{
  unsigned int x = g.x;
  signed int y = g.y >> 16;
  unsigned int c = g.piece[0];

  if(y >= -1)
  {
    g.b[y + 1][x] = ((c << 1) & (LU_COLOR_MASK | LU_BOMB)) | LU_COLOR_H;
    g.b[y + 1][x + 1] = ((c >> 1) & (LU_COLOR_MASK | LU_BOMB)) | LU_COLOR_H;
    if(y >= 0)
    {
      g.b[y][x] = ((c << 3) & (LU_COLOR_MASK | LU_BOMB)) | LU_COLOR_H;
      g.b[y][x + 1] = ((c << 5) & (LU_COLOR_MASK | LU_BOMB)) | LU_COLOR_H;
    }
    g.dirty_columns |= 3 << x;
    g.floaty_columns |= 3 << x;
    NOISECTRL = 0xf100;
    NOISEFREQ = FREQ_RESET | 0x70;
    return 0;
  }
  else
  {
    NOISECTRL = 0xf700;
    NOISEFREQ = FREQ_RESET | 0x80;
    return 1;
  }
}



/* pause() ******************
   Pauses the game. Returns 0 for continue or -1 for exit.
*/
static
int pause(void)
{
  unsigned char y_pos[4] = {48, 64, 96, 120};
  signed int result = -1;

  while(result < 0)
  {
    unsigned int x, y, c = 0x9380;

    for(y = 0; y < 4; y++)
      for(x = 64; x < 192; x += 32, c += 8)
      {
        u.oam[u.n_sprites].y = y_pos[y] | OAM_WIDE;
        u.oam[u.n_sprites].x = x | OAM_SIZE2;
        u.oam[u.n_sprites++].tile = c;
      }

    gba_draw_score();
    lu_wait4vbl(1);

    if(u.repeats[2] == 1)  /* select button: exit */
      result = 1;
    else if(u.repeats[3] == 1)  /* start button: resume */
      result = 0;

  }
  return result;
}


static
void lu_sweeping(void)
{
  u32 sweep_cd = u.sweep_sub + u.cur_song.music_speed;

  pf_gravity();

  mark_blawks();
  u.sweep_sub = sweep_cd & 0xffffff;
  if(sweep_cd >= 1<<24)
  {
    g.sweep_x++;
    if(g.sweep_x >= LU_FIELD_WID)
    {
      unsigned int score_add;

      if(g.skin_transition == LU_SKIN_WAIT_SONG_END)
        g.skin_transition = LU_SKIN_LOAD_SONG;

      sweep_finish();

      score_add = g.deletes_this_pass >= 4
                  ? (g.deletes_this_pass * 160)
                  : (g.deletes_this_pass * 40);
      if(score_add)
        add_score(8 * LU_SCR_LEFT + 12 * LU_FIELD_WID - 20,
                  21,
                  45,
                  score_add);
      g.sweep_x = 0;
      g.sweepgrp_left = 0;
      g.deletes_last_pass = g.deletes_this_pass;
      g.deletes_this_pass = 0;
      g.deletes_move_cd = g.deletes_last_pass ? 16 : 0;

      gsm_pass();
    }

    set_palette_dimmed(PALRAM+0xe0, PALRAM+0xa0, g.sweep_x & 1 ? RGB(31,31,31) : 0, 8);
    set_palette_dimmed(PALRAM+0xf0, PALRAM+0xb0, g.sweep_x & 1 ? RGB(31,31,31) : 0, 8);

    if(sweep_column(g.sweep_x))
    {
    }
    else
    {
      sweep_finish();

      g.sweepgrp_left = g.sweep_x + 1;
    }

    if((g.sweep_x == 0
        && g.n_pieces >= u.next_skin_at)
       || (SELECT_SWITCH_SKIN
           && !(JOY & JOY_SELECT)
           && g.skin_transition == LU_SKIN_IDLE))
    {
      u.next_skin_at += 100;
      g.skin_transition = LU_SKIN_LOAD_FILE;
    }
  }
}


static
void lu_block_motion(void)
{
  /* Left */
  if(u.repeats[5] == 1 || u.repeats[5] == REPEAT_DELAY)
  {
    if(!piece_overlaps(g.x - 1, (g.y + 0xf000) >> 16))
    {
      g.x--;
      SQR1CTRL = 0x8180;
      SQR1FREQ = FREQ_RESET | u.cur_song.move_period;
    }
  }

  /* Right */
  if(u.repeats[4] == 1 || u.repeats[4] == REPEAT_DELAY)
  {
    if(!piece_overlaps(g.x + 1, (g.y + 0xf000) >> 16))
    {
      g.x++;
      SQR1CTRL = 0x8180;
      SQR1FREQ = FREQ_RESET | u.cur_song.move_period;
    }
  }

  /* Fast left */
  if(u.repeats[9] == 1)
  {
    while(!piece_overlaps(g.x - 1, (g.y + 0xf000) >> 16))
      g.x--;
  }

  /* Slam right */
  if(u.repeats[8] == 1)
  {
    while(!piece_overlaps(g.x + 1, (g.y + 0xf000) >> 16))
      g.x++;
  }


  /* Rotate Left */
  if(u.repeats[1] == 1)
  {
    g.piece[0] = ((g.piece[0] << 2) | (g.piece[0] >> 6)) & 0xff;
    SQR1CTRL = 0x8180;
    SQR1FREQ = FREQ_RESET | u.cur_song.turn_period;
  }

  /* Rotate Right */
  if(u.repeats[0] == 1)
  {
    g.piece[0] = ((g.piece[0] << 6) | (g.piece[0] >> 2)) & 0xff;
    SQR1CTRL = 0x8180;
    SQR1FREQ = FREQ_RESET | u.cur_song.turn_period;
  }

  /* Soft Drop */
  if(u.repeats[7] == 1)
    u.state_time = 0;
  else if(u.state_time > 0)
    u.state_time--;
  if(u.state_time == 0)
  {
    if(u.repeats[7])
    {
      g.y += 0xff00;
      g.score++;
    }
    else
      g.y += g.y_speed;

    if(piece_overlaps(g.x, (g.y + 0xf000) >> 16))
    {
      g.y -= 0x1000;
      if(lock_piece())
      {
        game_over();
      }
      else
        make_new_piece();
    }
  }
}


void Lplay(void)
{
  char done = 0;

  /* test drawing */
  init_vram();
  clear_keybuf();
  u.state = LU_NEW_GAME;
  u.paused = 0;
  g.score = 0;
  g.cur_skin = -1;

  /* test sweeping */
  while(!done)
  {
    if(g.deletes_move_cd)
    {
      if(--g.deletes_move_cd == 0)
      {
        g.deletes += g.deletes_last_pass;
      }
    }
    switch(u.state)
    {
    case LU_NEW_GAME:
      new_game();
      break;
    case LU_NEW_PIECE:
      lu_sweeping();
      if(--u.state_time == 0)
      {
        u.state_time = 0x60000 / g.y_speed;
        u.state = LU_FALLING_PIECE;
      }
      break;
    case LU_FALLING_PIECE:
      lu_sweeping();
      lu_block_motion();
    default:
      break;
    }

    rest_frames(1);
    handle_bonuses();
    if(u.repeats[3] == 1)
    {
      done = pause();
      if(u.state == LU_GAMEOVER)
        u.state = LU_NEW_GAME;
    }
  }
  DMA[0].control = 0;
  LCDMODE = 0;
  PALRAM[0] = RGB(0, 0, 0);
}
