/* mine.c
   minesweeper clone

Copyright 2005 Damian Yerrick

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Author contact information is in readme.txt

*/

#include "mine.h"

#define SHOW_TILES_WITH_NO_UNKNOWN_NEIGHBORS 0
#define PROFILE 0

#define g game_globals.right


const u16 basic_pal[16] =
{
  RGB(27,27,27),RGB( 0, 0,23),RGB( 0,23, 0),RGB( 0,23,23),
  RGB(23, 0, 0),RGB(23, 0,23),RGB(23,23, 0),RGB(21,21,21),
  RGB( 0, 0, 0),RGB( 0, 0,31),RGB( 0,31, 0),RGB( 0,31,31),
  RGB(31, 0, 0),RGB(31, 0,31),RGB(31,31, 0),RGB(31,31,31)
};

static void draw_row_small(unsigned int row)
{
  const unsigned int ormask = 0x0010;
  unsigned short *dst = MAP[PLAYFIELD_MAP][row + 4];
  unsigned char *src = g.pf[row];
  unsigned int wid;

  for(wid = PLAYFIELD_WID; wid > 0; wid--)
  {
    unsigned int cur_space = *src++;

#if SHOW_TILES_WITH_NO_UNKNOWN_NEIGHBORS
    if(cur_space & 0x40)
      *dst++ = '!';
    else
#endif
      *dst++ = (cur_space & 0x0f) | ormask;
  }
}


static inline void draw_row(unsigned int row)
{
  draw_row_small(row);
}  


void draw_dirty_rows(void)
{
  unsigned int shift_reg;
  unsigned int y;

  for(y = 0, shift_reg = g.dirty_rows;
      y < PLAYFIELD_HT && shift_reg != 0;
      y++, shift_reg >>= 1)
  {
    if(shift_reg & 1)
      draw_row(y);
  }
  g.dirty_rows = 0;
}



/* gen_map() ***************************
   Generate a 30x16 square map with n_mines mines, none of them at
   position (base_x, base_y).
*/
void gen_map(unsigned int base_x, unsigned int base_y, unsigned int n_mines)
{
  unsigned int x, y;

  for(x = 0; x < PLAYFIELD_WID; x++)
    g.reveal_yp2[x] = 0;
  for(x = 0; x < PLAYFIELD_WID; x++)
    g.reveal_yp1[x] = 0;

  for(y = 0; y < PLAYFIELD_HT; y++)
    for(x = 0; x < PLAYFIELD_WID; x++)
      g.pf[y][x] = 0x0d;
  for(y = g.pf_top; y < g.pf_bottom; y++)
    for(x = g.pf_left; x < g.pf_right; x++)
      g.pf[y][x] = 0x0f;
  g.n_mines = n_mines;

  while(n_mines > 0)
  {
    /* Can't put a mine off the playing field, in a corner,
       under the starting cursor, or on another mine. */
    x = (rand() >> 8) & 0x1f;
    if(x < g.pf_left || x >= g.pf_right)  /* off left or right sides */
      continue;
    y = (rand() >> 8) & 0x0f;
    if(y < g.pf_top || y >= g.pf_bottom)  /* off top or bottom sides */
      continue;
    if(g.pf[y][x] != 0x0f)  /* already a mine there */
      continue;
    if(x == base_x && y == base_y)  /* under cursor */
      continue;
    if((x == g.pf_left || x == g.pf_right - 1)
       && (y == g.pf_top || y == g.pf_bottom - 1))  /* in a corner */
      continue;

    g.pf[y][x] = 0x8f;
    n_mines--;
  }
  g.dirty_rows = (1 << PLAYFIELD_HT) - 1;
}


/* touch_space() ***********************
   Touches a space.  Returns the content of the space
   (0-8: number; 9: mine)
*/
unsigned int touch_space(unsigned int x_origin, unsigned int y_origin)
{
  unsigned int left = g.pf_left, top = g.pf_top;
  unsigned int right = g.pf_right, bottom = g.pf_bottom;
  unsigned int mines = 0;
  unsigned int x, y;

  if(g.pf[y_origin][x_origin] & 0x80)
  {
    g.pf[y_origin][x_origin] = 0x89;
    return 9;
  }

  if(x_origin > left)
    left = x_origin - 1;
  if(x_origin < right - 1)
    right = x_origin + 2;
  if(y_origin > top)
    top = y_origin - 1;
  if(y_origin < bottom - 1)
    bottom = y_origin + 2;

  for(y = top; y < bottom; y++)
    for(x = left; x < right; x++)
      if(g.pf[y][x] & 0x80)
        mines++;

  g.pf[y_origin][x_origin] = mines;
  g.dirty_rows |= 1 << y_origin;

  return mines;
}


/* count_mines_unknowns() **************
   Counts the number of proven mines and unknown spaces around a
   given space.
   unks_out: If not NULL, then number of unknown spaces (blank or
   flagged) is put here.
   flags_out: If not NULL, then number of flagged spaces is
   put here.
   return value: number of proven mines
*/
static
unsigned int count_mines_unknowns(unsigned int x_origin,
                                  unsigned int y_origin,
                                  unsigned int *unks_out,
                                  unsigned int *flags_out)
{
  unsigned int left = g.pf_left, top = g.pf_top;
  unsigned int right = g.pf_right, bottom = g.pf_bottom;
  unsigned int mines = 0, unks = 0, flags = 0;
  unsigned int x, y;

  if(x_origin > left)
    left = x_origin - 1;
  if(x_origin < right - 1)
    right = x_origin + 2;
  if(y_origin > top)
    top = y_origin - 1;
  if(y_origin < bottom - 1)
    bottom = y_origin + 2;

  for(y = top; y < bottom; y++)
    for(x = left; x < right; x++)
      switch(g.pf[y][x] & 0x0f)
      {
      case 9:
        mines++;
        break;
      case 14:
        flags++;
        /* fall through */
      case 15:
        unks++;
        break;
      }

  if(unks_out)
    *unks_out = unks;
  if(flags_out)
    *flags_out = flags;

  return mines;
}



/* touch_around() ************
   If a revealed space has one or more mines touching it, and the
   correct number of flags surrounds a space, then touch all unknown
   spaces surrounding the space.
   Return 0 for no squares opened or nonzero for at least one square opened.
*/
int touch_around(unsigned int x_origin, unsigned int y_origin)
{
  unsigned int unks, mines;
  unsigned int cur_space = g.pf[y_origin][x_origin] & 0x0f;
  int any = 0;

  if(cur_space < 1 || cur_space > 8)
    return 0;

  mines = count_mines_unknowns(x_origin, y_origin, &unks, NULL);

  if(unks > 0 && mines == cur_space)
  {
    unsigned int left = x_origin > g.pf_left ? x_origin - 1 : x_origin;
    unsigned int right = x_origin < (g.pf_right - 1)
                         ? x_origin + 2
                         : x_origin + 1;
    unsigned int top = y_origin > g.pf_top ? y_origin - 1 : y_origin;
    unsigned int bottom = y_origin < (g.pf_bottom - 1)
                          ? y_origin + 2
                          : y_origin + 1;
    unsigned int x, y;

    for(y = top; y < bottom; y++)
      for(x = left; x < right; x++)
        if((g.pf[y][x] & 0x0f) == 0x0f)
        {
          touch_space(x, y);
          any++;
        }
  }
  return any;
}


static
void B_action(unsigned int x, unsigned int y)
{
  unsigned int cur_space = g.pf[y][x];

  /* if an unknown, toggle its flag */
  if((cur_space & 0x0e) == 0x0e)
  {
    g.pf[y][x] = cur_space ^ 1;
    g.dirty_rows |= 1 << y;
  }
}


static
int A_action(unsigned int x, unsigned int y)
{
  unsigned int cur_space = g.pf[y][x] & 0x0f;

  if(cur_space == 15)  /* touch only unknown unflagged spaces */
    return touch_space(x, y);
  else
    return -1;  /* nothing touched */
}


/* AB_action() ************************
   If number of proven mines and flags around a space equals
   number of mines, touch all the unknowns around the space.
   Returns -1 for done nothing, 0 for ok, or >0 for game over.
*/
static
int AB_action(unsigned int x_origin, unsigned int y_origin)
{
  unsigned int flags, mines, unks;
  unsigned int cur_space = g.pf[y_origin][x_origin] & 0x0f;

  /* only accept spaces marked 1 through 8 */
  if(cur_space == 0 || cur_space >= 9)
    return -1;

  mines = count_mines_unknowns(x_origin, y_origin, &unks, &flags);
  if(unks > 0 && mines + flags == cur_space)
  {
    unsigned int left = x_origin > g.pf_left ? x_origin - 1 : x_origin;
    unsigned int right = x_origin < (g.pf_right - 1)
                         ? x_origin + 2
                         : x_origin + 1;
    unsigned int top = y_origin > g.pf_top ? y_origin - 1 : y_origin;
    unsigned int bottom = y_origin < (g.pf_bottom - 1)
                          ? y_origin + 2
                          : y_origin + 1;
    unsigned int x, y;

    mines = 0;

    for(y = top; y < bottom; y++)
      for(x = left; x < right; x++)
        if((g.pf[y][x] & 0x0f) == 0x0f)
        {
          if(touch_space(x, y) == 9)
            mines++;
        }
    return mines ? 9 : cur_space;
  }
  else
    return -1;
}




/* scan_board() ************************
   Scan the playfield for provable mines and non-mines.
   Uses the Single Point strategy as defined in
http://www.ccs.neu.edu/home/ramsdell/pgms/doc/sp.SinglePointStrategy.html
*/
static
int scan_board(void)
{
  unsigned char pf[PLAYFIELD_HT][32] = {{0}};
  unsigned int x, y;
  unsigned int mines, unks, cur_space;
  unsigned int any = 0;
#if PROFILE
  unsigned int bgc = PALRAM[0];

  PALRAM[0] = RGB(31,31, 0);
#endif

  for(y = g.pf_top; y < g.pf_bottom; y++)
  {
    for(x = g.pf_left; x < g.pf_right; x++)
    {
      /* 0: do nothing; 1: flag as mine; 2: touch */
      unsigned int all = 0;

      cur_space = g.pf[y][x] & 0x7f;

      /* If this is a mine (0x09), space touching no unknown spaces (0x40),
         or an unknown space (0x0e or 0x0f), skip. */
      if(cur_space >= 0x09)
        continue;

      /* Proofs always start from a space adjacent to a known number
         of mines and at least one unknown. */
      mines = count_mines_unknowns(x, y, &unks, NULL);
      if(unks == 0)
      {
        /* If there are no unknowns, mark this space as touching no unknown
           spaces so that we know to skip it next time. */
        g.pf[y][x] = cur_space | 0x40;
        continue;
      }

      if(cur_space == 0)
        all = 2;

#if 0  /* used in single-point solver, not sweep mode */
      /* Autoflag mines.
         If a space has n mines by it, and n (proven mines + unknowns)
         surround the space, then all unknowns are proven mines.
      */
      else if(0 && cur_space >= 1 && cur_space <= 8
           && mines + unks == cur_space)
          all = 1;

      /* Prove spaces.
         If a space has n mines by it, all proven, and > 0 unknowns,
         then none of the unknowns contains a mine.  Click them all.
      */
      else if(mines == cur_space)
        all = 2;
#endif
      if(all)
      {
        unsigned int left = g.pf_left, top = g.pf_top;
        unsigned int right = g.pf_right, bottom = g.pf_bottom;
        unsigned int xc, yc;

        any = 1;

        if(x > left)
          left = x - 1;
        if(x < right - 1)
          right = x + 2;
        if(y > top)
          top = y - 1;
        if(y < bottom - 1)
          bottom = y + 2;

        for(yc = top; yc < bottom; yc++)
          for(xc = left; xc < right; xc++)
            if((g.pf[yc][xc] & 0x0e) == 0x0e)  /* 14: flag; 15: empty */
              pf[yc][xc] = all;  /* mark for proving */
      }
    }
  }
#if PROFILE
  PALRAM[0] = RGB(31,31,31);
#endif

  /* if there's anything to click on this pass, make it so */
  if(any)
    for(y = 0; y < PLAYFIELD_HT; y++)
      for(x = 0; x < PLAYFIELD_WID; x++)
        switch(pf[y][x])
        {
        case 1:
          g.pf[y][x] = 0x89;
          g.n_mines--;
          g.dirty_rows |= 1 << y;
          break;
        case 2:
          touch_space(x, y);
          break;
        }
#if PROFILE
  PALRAM[0] = bgc;
#endif

  return any;
}


static
const char hexdigits[16] = "0123456789ABCDEF";



void sweep_board(unsigned int row)
{
  unsigned char reveal_y[32] = {0};
  unsigned int x, y;
  unsigned int mines, unks, cur_space;
  unsigned int any = 0;
#if PROFILE
  unsigned int bgc = PALRAM[0];

  PALRAM[0] = RGB(31,31, 0);
#endif

  /* If the sweeper is on the LAST row, evaluate the FIRST row. */
  if(row >= g.pf_bottom - 1 || row < g.pf_top)
  {
    y = g.pf_top;
    for(x = 0; x < PLAYFIELD_WID; x++)
      g.reveal_yp2[x] = 0;
  }
  else
    y = row + 1;

#if 0
  MAP[PLAYFIELD_MAP][0][20] = hexdigits[row];
  MAP[PLAYFIELD_MAP][0][21] = hexdigits[y];
#endif

  if(y >= g.pf_bottom || y < g.pf_top)
    return;

  for(x = g.pf_left; x < g.pf_right; x++)
  {
    reveal_y[x] = g.reveal_yp1[x];
    g.reveal_yp1[x] = g.reveal_yp2[x];
    g.reveal_yp2[x] = 0;
  }

  for(x = g.pf_left; x < g.pf_right; x++)
  {
    /* 0: do nothing; 1: flag as mine; 2: touch */
    unsigned int all = 0;

    cur_space = g.pf[y][x] & 0x7f;

    /* Skip mines, unknown spaces, and spaces touching no unknown spaces. */
    if(cur_space >= 0x09)
      continue;

    /* Proofs always start from a space adjacent to a known number
       of mines and at least one unknown. */
    mines = count_mines_unknowns(x, y, &unks, NULL);
    if(unks == 0)
    {
      /* If there are no unknowns, mark this space as touching no unknown
         spaces so that we know to skip it next time. */
      g.pf[y][x] = cur_space | 0x40;
      continue;
    }

    if(cur_space == 0)
      all = 2;

    /* Autoflag mines. */
    else if(cur_space >= 1 && cur_space <= 8
         && mines + unks == cur_space)
      all = 1;

    /* Prove spaces. */
    else if(mines == cur_space)
      all = 2;

    if(all)
    {
      unsigned int left = g.pf_left, right = g.pf_right;
      unsigned int xc;

      any = 1;

      if(x > left)
        left = x - 1;
      if(x < right - 1)
        right = x + 2;

      if(y > g.pf_top)  /* row above */
        for(xc = left; xc < right; xc++)
          if((g.pf[y - 1][xc] & 0x0e) == 0x0e)  /* 14: flag; 15: empty */
            reveal_y[xc] = all;

      for(xc = left; xc < right; xc++)
        if((g.pf[y][xc] & 0x0e) == 0x0e)  /* 14: flag; 15: empty */
          g.reveal_yp1[xc] = all;

      if(y < g.pf_bottom - 1)
        for(xc = left; xc < right; xc++)
          if((g.pf[y + 1][xc] & 0x0e) == 0x0e)  /* 14: flag; 15: empty */
            g.reveal_yp2[xc] = all;
    }
  }

  /* and reveal them */
  if(row >= g.pf_top && row < g.pf_bottom)
    for(x = g.pf_left; x < g.pf_right; x++)
      switch(reveal_y[x])
      {
      case 1:
        g.pf[row][x] = 0x89;
        g.n_mines--;
        g.dirty_rows |= 1 << row;
        break;
      case 2:
        touch_space(x, row);
        break;
      }
#if PROFILE
  PALRAM[0] = bgc;
#endif
}



/* gameoverify() ******************
   Show all mines on the board, for game over.
*/
static
void gameoverify(void)
{
  unsigned int x, y;
  for(y = g.pf_top; y < g.pf_bottom; y++)
  {
    for(x = 0; x < PLAYFIELD_WID; x++)
    {
      if(g.pf[y][x] & 0x80)
      {
        g.pf[y][x] = 0x8a;
        g.dirty_rows |= 1 << y;
      }
    }
  }
}


/* debug_verify_count() ***********************
   Verify that no mines were inadvertently lost.
*/
static
int debug_verify_count(void)
{
  unsigned int n_mines = 0;
  unsigned int x, y;

  for(y = 0; y < PLAYFIELD_HT; y++)
    for(x = 0; x < PLAYFIELD_WID; x++)
    {
      int c = g.pf[y][x];

      if((c & 0x80) && (c != 0x89))
        n_mines++;
    }

  return n_mines - g.n_mines;
}


static
void draw_hud(unsigned int x, unsigned int y)
{
  OAM[0].y = (((y << 3) + 28) & 0xff) | OAM_16C | OAM_SQUARE;
  OAM[0].x = (((x << 3) - 4) & 0x1ff) | OAM_SIZE1;
  OAM[0].tile = 0x0c;
}


static
inline void cputs(unsigned int x, unsigned int y, const char *str)
{
  nttextout(PLAYFIELD_MAP, x, y, 0, str);
}


static
void init_vram(void)
{
  static const unsigned char gray_pal[4] = {0,8,7,15};
  static const unsigned shortsweeper_pal[8] =
  {
    RGB( 0, 0, 0),
    RGB( 8, 0, 0),
    RGB(16, 0, 0),
    RGB(24, 0, 0),
    RGB(31, 0, 0),
    RGB(31,15, 0),
    RGB(31,31, 0),
    RGB(31,31,31)
  };
  unsigned int len;
  const void *src;
  u32 *dst;

  while(LCD_Y != 161) ;
  LCDMODE = LCDMODE_BLANK;
  dma_memcpy(PALRAM +   0, basic_pal, sizeof(basic_pal));
  dma_memcpy(PALRAM + 256, basic_pal, sizeof(basic_pal));

  src = gbfs_get_obj(data_gbfs, "Rhud.chr", &len);
  expand_gb_to_gba4bit(PATRAM4(0, 0), src, len, gray_pal);
  expand_gb_to_gba4bit(SPR_VRAM(0), src, len, gray_pal);

  src = gbfs_get_obj(data_gbfs, "Rblocks.chr", &len);
  dma_memcpy(PATRAM4(0, 16), src, 16*32);
  dma_memcpy(PATRAM4(0, 96), (unsigned char *)src + 16*32, 32*32);

  for(len = 0; len < 128; len++)
  {
    OAM[len].y = OAM_HIDDEN;
  }

  for(len = 0; len < 32*20; len++)
  {
    MAP[PLAYFIELD_MAP][0][len] = ' ';
  }


  dst = PATRAM4(0, 7);
  for(len = 0; len < 8; len++)
  {
    unsigned int c = len | (len << 4);
    c |= c << 8;
    c |= c << 16;
    *dst++ = c;
  }

  dst = (u32 *)MAP[SWEEPER_MAP][1];
  for(len = 0; len < 512; len++)
    *dst++ = 0;

  for(len = 0; len < 30; len++)
    MAP[SWEEPER_MAP][15][len] = 0x1007;
  for(len = 0; len < 30; len++)
    MAP[SWEEPER_MAP][31][len] = 0x1007;

  for(len = 0; len < 8; len++)
    PALRAM[len + 16] = shortsweeper_pal[len];

  while(LCD_Y != 200) ;
  BGCTRL[0] = BGCTRL_PAT(0) | BGCTRL_16C | BGCTRL_NAME(SWEEPER_MAP)
              | BGCTRL_H32 | BGCTRL_V32;
  BGCTRL[2] = BGCTRL_PAT(0) | BGCTRL_16C | BGCTRL_NAME(PLAYFIELD_MAP)
              | BGCTRL_H32 | BGCTRL_V32;
  BGSCROLL[0].x = 0;
  BGSCROLL[0].y = 0;
  BGSCROLL[2].x = 0;
  BGSCROLL[2].y = 0;
  LCDMODE = 0 | LCDMODE_BG2 | LCDMODE_SPR | LCDMODE_1DSPR;
  BLENDMODE = BLENDMODE_NORMAL
              | BLENDMODE_BG0TOP
              | BLENDMODE_BG2BOT | BLENDMODE_SKYBOT;
  BLENDFRAC = BLENDFRAC_TOP(16) | BLENDFRAC_BOT(16);

  /* set a window, and show only the status bar (hide the sweeper) inside it */
  WIN_X[0] = WIN_DIM(0, 240);
  WIN_Y[0] = WIN_DIM(0, 32);
  WININ = WININ_W0BG2 | WININ_W0SPR | WININ_OUTALL;
  
  cputs( 1, 1, "B\031MBS");
  cputs(24, 1, "\010TIME");
}


/* cputn() *******
   converts a number n to decimal, outputting the last len digits
   with the rightmost digit at (x, y), using the lcd font
*/
void cputn(unsigned int n, unsigned int x, unsigned int y, unsigned int len)
{
  for(; len > 0; len--, x--)
  {
    unsigned int div_10 = dv(n, 10);
    unsigned int digit = n - 10 * div_10;

    MAP[PLAYFIELD_MAP][y][x] = digit + 0x60;
    MAP[PLAYFIELD_MAP][y + 1][x] = digit + 0x70;
    n = div_10;
  }
}


void Rplay(void)
{
  unsigned int curs_x = 14, curs_y = 7;
  char need_new_map = 1, need_scan = 0, need_touch_around = 0, moved = 0;
  unsigned int sweep_y = 0;
  unsigned char tix = 0;
  unsigned int seconds = 0;
  

  init_vram();
  g.pf_left = 0;
  g.pf_right = 30;
  g.pf_top = 0;
  g.pf_bottom = 16;
  g.dirty_rows = 0;
  g.n_mines = 0;

  u.sweep_sub = 0;

  cputs( 1, 4, "RIGHT MODE    BUILD 20050618");

  cputs( 1, 6, "PRESS A TO PLAY");
  cputs( 1, 7, "PRESS B TO RETURN TO MENU");

  cputs( 1,10, "CONTROLS:");
  cputs( 1,11, "D-PAD - MOVE CURSOR");
  cputs( 1,12, "A     - UNCOVER SPACE");
  cputs( 1,13, "B     - FLAG B\031MB LOCATION");
  cputs( 1,14, "A+B   - TOUCH UNFLAGGED");
  cputs( 1,15, "        SPACES AROUND CURSOR");
  cputs( 1,17, "NUMBER TELLS YOU HOW MANY");
  cputs( 1,18, "B\031MBS TOUCH THE SPACE.");
  cputs( 1,19, "DON'T STEP ON ONE!");

  while(need_new_map != 62)
  {
    make_repeats(JOY ^ 0x3ff);

    if(u.repeats[4] == 1 || u.repeats[4] == REPEAT_DELAY)  /* RIGHT */
    {
      curs_x++;
      if(curs_x >= g.pf_right)
        curs_x = g.pf_left;
      moved = 1;
      need_touch_around = 1;
    }
    if(u.repeats[5] == 1 || u.repeats[5] == REPEAT_DELAY)  /* LEFT */
    {
      if(curs_x <= g.pf_left)
        curs_x = g.pf_right - 1;
      else
        curs_x--;
      moved = 1;
      need_touch_around = 1;
    }
    if(u.repeats[6] == 1 || u.repeats[6] == REPEAT_DELAY)  /* UP */
    {
      if(curs_y == g.pf_top)
        curs_y = g.pf_bottom - 1;
      else
        curs_y--;
      moved = 1;
      need_touch_around = 1;
    }
    if(u.repeats[7] == 1 || u.repeats[7] == REPEAT_DELAY)  /* DOWN */
    {
      curs_y++;
      if(curs_y >= g.pf_bottom)
        curs_y = g.pf_top;
      moved = 1;
      need_touch_around = 1;
    }

    if(u.repeats[0])  /* A */
    {
      if(need_new_map == 15 && u.repeats[0] == 1)
      {
        static const unsigned char widths[4] = {16, 30, 9, 8};
        static const unsigned char heights[4] = {16, 16, 9, 8};
        static const unsigned char mines[4] = {40, 99, 10, 10};
        unsigned int lvl_type = ((JOY ^ 0x3ff) >> 8) & 3;
        unsigned int w = widths[lvl_type];
        unsigned int h = heights[lvl_type];

        g.pf_left = (30 - w) / 2;
        g.pf_right = g.pf_left + w;
        g.pf_top = (16 - h) / 2;
        g.pf_bottom = g.pf_top + h;

        if(curs_x < g.pf_left || curs_x >= g.pf_right)
          curs_x = 14;
        if(curs_y < g.pf_top || curs_y >= g.pf_bottom)
          curs_y = 7;

        /* show sweeper */
        LCDMODE = 0 | LCDMODE_BG0 | LCDMODE_BG2 | LCDMODE_SPR
                    | LCDMODE_1DSPR | LCDMODE_WIN0;
        PALRAM[0] = basic_pal[0];
        srand(rand() ^ tix);
        srand(rand() ^ seconds);
        gen_map(curs_x, curs_y, mines[lvl_type]);
        need_new_map = 0;
        u.sweep_sub = 0;
        sweep_y = 8; /* should be 15 */
        tix = 0;
        seconds = 0;
        moved = 1;
      }
      if(!need_new_map)
      {
        int a = -1;

        if(u.repeats[1] == 1
           || (u.repeats[0] == 1 && u.repeats[1])
           || (u.repeats[1] && moved))
          a = AB_action(curs_x, curs_y);
        else if(moved)
          a = A_action(curs_x, curs_y);

        if(a == 9)
        {
          PALRAM[0] = RGB(31, 0, 0);
          gameoverify();
          need_new_map = 1;
        }
        else if(a >= 0)
        {
          need_scan = 1;
          need_touch_around = 1;
        }
        moved = 0;
      }
    }
    if(u.repeats[1] == 1)  /* B */
    {
      moved = 1;
      if(need_new_map >= 15)
        need_new_map = 62;
      else
        B_action(curs_x, curs_y);
    }

    if(need_touch_around)
    {
      need_scan |= touch_around(curs_x, curs_y);
      need_touch_around = 0;
    }

#if PROFILE
    {
      unsigned int lcd_y = LCD_Y + 68;

      if(lcd_y > 228)
        lcd_y -= 228;

      cputn(lcd_y, 15, 0, 2);
    }        
#endif

    u.retraces++;
    wait4vbl();
    draw_dirty_rows();
    draw_hud(curs_x, curs_y);
    cputn(g.n_mines, 4, 2, 3);

    if(need_new_map > 0 && need_new_map < 15)
      need_new_map++;

    if(!need_new_map)
    {
      ++tix;
      /* assume a 59.75 Hz game loop, which is close to
         the GBA's 59.7275 Hz */
      if(tix == 1 || tix == 60 || tix == 120 || tix == 180)
      {
        seconds++;
        cputn(seconds, 27, 2, 3);
      }
      if(tix == 239)
        tix = 0;

      if(need_scan)
      {
        need_scan = scan_board();
        need_touch_around |= need_scan;
      }

      u.sweep_sub += 65536 * 18;
      BGSCROLL[0].x = 0;
      BGSCROLL[0].y = 226 - sweep_y * 8 - ((int)u.sweep_sub >> 21);
      if(u.sweep_sub >= 1<<24)
      {
        sweep_board(sweep_y);
        need_touch_around = need_scan = 1;
        sweep_y++;
        if(sweep_y >= PLAYFIELD_HT)
        {
          sweep_y = 0;
        }
        u.sweep_sub -= 1<<24;
      }
      else if(g.n_mines == 0)
      {
        PALRAM[0] = RGB( 0,31, 0);
        need_new_map = 1;
      }
      else if(debug_verify_count())
      {
        PALRAM[0] = RGB(0, 0, 31);
        need_new_map = 1;
      }
      else
        PALRAM[0] = basic_pal[0];
    }
  }
  BLENDMODE = 0;
}
