/* util.c
   utilities for luminesweeper

Copyright 2005 Damian Yerrick

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Author contact information is in readme.txt

*/

#include "mine.h"

UtilGlobals u;

/* clear_keybuf() **********************
   Clear the autorepeats.
*/
void clear_keybuf(void)
{
  int i;

  for(i = 0; i < 10; i++)
    u.repeats[i] = 1;
}


/* make_repeats() **********************
   Handle auto-repeat for ten-bit joypad.
*/
void make_repeats(u32 keys)
{
  unsigned int i;

  for(i = 0;
      i < 10;
      keys >>= 1, i++)
  {
    if(keys & 1)
    {
      u.repeats[i]++;
      if(u.repeats[i] >= REPEAT_DELAY + 2)
        u.repeats[i] = REPEAT_DELAY;
    }
    else
      u.repeats[i] = 0;
  }
}


/* dma_memcpy() ************************
   Copy a region of memory to VRAM or any other area of memory
   using the GBA hardware.  It needs a 2-cycle wait after writing
   to the control register; making this a non-inline function
   introduces enough rest time (bx lr).
*/
void dma_memcpy(void *dst, const void *src, size_t len)
{
  DMA[3].src = src;
  DMA[3].dst = dst;
  DMA[3].count = len >> 2;
  DMA[3].control = DMA_ENABLE | DMA_COPYNOW | DMA_U32 | DMA_SRCINC | DMA_DSTINC;
}


/* nttextout() *************************
   Write the given text to a nametable.
*/
void nttextout(u32 nt, u32 x, u32 y, u32 c, const char *str)
{
  while(*str)
    MAP[nt][y][x++] = (*str++ & 0xff) | c;
}


#if 0
/* itoa_lpad() *************************
   Convert n into a string of len characters, left-padded with
   lpad_chr.  buf points to a buffer of at least (len + 1) chars.
*/
void itoa_lpad(char *buf, size_t len, unsigned int n, int lpad_chr)
{
  buf[len] = 0;
  if(n == 0)
    n = 0;

  /* extract each digit */
  do {
    /* assumes compiler can optimize n % 10 and n / 10 into one op
       also assumes digits appear consecutively in 0123456789 order
       (this is true of ascii) */
    buf[--len] = (n % 10) + '0';
    n = n / 10;
  } while(n && len);

  /* left pad */
  while(len)
    buf[--len] = lpad_chr;
}
#endif


/* expand_gb_to_gba4bit() *********
   Expand tiles from Game Boy format to Game Boy Advance
   format in 4-bits
*/
void expand_gb_to_gba4bit(void *in_dst, const u8 *src, size_t len, const unsigned char gbpal[4])
{
  u32 *dst = in_dst;
  const unsigned char default_gbpal[4] = {0, 1, 2, 3};

  if(!gbpal)
    gbpal = default_gbpal;

  len >>= 1;  /* convert from bytes to scanlines */

  for(; len > 0; len--)
  {
    unsigned int sr0 = *src++;
    unsigned int sr1 = *src++;
    unsigned int dstbits = 0;
    unsigned int bits = 8;

    for(; bits > 0; bits--)
    {
      unsigned int gb_color = ((sr1 & 1) << 1) | (sr0 & 1);
      sr0 >>= 1;
      sr1 >>= 1;
      dstbits = (dstbits << 4) | gbpal[gb_color];
    }
    *dst++ = dstbits;
  }
}


void set_palette_dimmed(void *in_dst,
                        const void *in_src,
                        unsigned int dim_to,
                        unsigned int dim_level)
{
  u32 *dst = in_dst;
  const u32 *src = in_src;
  unsigned int n;
  unsigned int to_expanded = (dim_to & 0xFFFF) | (dim_to << 16);
  unsigned int to_0 = ((to_expanded & 0x7C1F03E0) >> 4) * dim_level;
  unsigned int to_1 = ((to_expanded & 0x03E07C1F)) * dim_level;

  /* optimization for fade all the way out */
  if(dim_level == 0)
  {
    dma_memcpy(in_dst, in_src, 32);
    return;
  }

  dim_level = 16 - dim_level;

  for(n = 8; n > 0; n--)
  {
    unsigned int from = *src++;
    unsigned int from_0 = ((from & 0x7C1F03E0) >> 4) * dim_level;
    unsigned int from_1 = ((from & 0x03E07C1F)) * dim_level;

    from_0 = ((from_0 + to_0)) & 0x7C1F03E0;
    from_1 = ((from_1 + to_1) >> 4) & 0x03E07C1F;

    *dst++ = from_0 | from_1;
  }
}
